export const DIRECTIVE_MAP = {
	'=': 'equals',
	'!=': 'not-equals',
	'>=': 'greater-than-equal',
	'<=': 'less-than-equal',
	'>': 'greater-than',
	'<': 'less-than',
};

export const DATA_SOURCE_QUERY_OPERATORS = Object.keys(DIRECTIVE_MAP) as (keyof typeof DIRECTIVE_MAP)[];

export type DataSourceQueryOperator = typeof DATA_SOURCE_QUERY_OPERATORS[number];

export interface DataSourceOperatorQuery {
	lhs: string;
	operator: DataSourceQueryOperator;
	rhs: string;
}

export interface DataSourceFunctionQuery {
	name: string;
	args: string[];
}

export type DataSourceQuery = DataSourceOperatorQuery | DataSourceFunctionQuery;

interface ConvertToDataSourceQueryCollectionOptions {
}

const parseDirectives = (valueRaw: string) => {
	const fragments = valueRaw.split('.');

	if (!(
		fragments.length > 1 && fragments[0].length === 0
	)) {
		return {
			operator: '=',
			rhs: valueRaw,
		};
	}

	const theOperator = fragments.slice(1, -1).reduce(
		(theState, f) => {
			const whichDirective = Object.entries(DIRECTIVE_MAP).find(
				([, match]) => {
					const remainingChars = match.split('').reduce(
						(rem, matchChar) => {
							if (matchChar === rem.charAt(0)) {
								return rem.slice(1);
							}

							return rem;
						},
						f.toLowerCase()
					);

					return remainingChars.length <= 0;
				},
			);

			const selectedOperator = whichDirective?.[0] ?? theState.operator;

			if (selectedOperator === '=') {
				if (theState.operator === '>') {
					return {
						...theState,
						operator: '>=',
						currentToken: `${theState.currentToken}${f}`,
					};
				}

				if (theState.operator === '<') {
					return {
						...theState,
						operator: '<=',
						currentToken: `${theState.currentToken}${f}`,
					};
				}
			}

			return {
				...theState,
				operator: selectedOperator,
				currentToken: `${theState.currentToken}${f}`,
			};
		},
		{
			operator: undefined as (string | undefined),
			currentToken: '',
		},
	);

	return {
		operator: theOperator.operator ?? '=',
		rhs: fragments.at(-1),
	};
};

export const convertToDataSourceQueryCollection = (
	q: URLSearchParams,
	options = {} as ConvertToDataSourceQueryCollectionOptions,
): DataSourceQuery[] => {
	return Array.from(q.entries())
		.reduce(
			(queries, [key, value]) => {
				const s = parseDirectives(value);

				return [
					...queries,
					{
						lhs: key,
						operator: s.operator,
						rhs: JSON.stringify(s.rhs),
					},
				];
			},
			[] as DataSourceQuery[]
		);
	// all the queries are to be treated as "AND", as suggested by the & character to separate the query param entries
};
